C----------------------------------------------------------------
C---------------------------------------------------------------------
C               Module *** GPOI_F ***   A.Filhol 23/07/91
C
C    Generation de nombres aleatoires selon une loi de POISSON de
C    parametre MT. (Version FORTRAN 77)
C
C    J.Berruyer, A.Antoniadis, A.Filhol  (1985)  Rapport ILL 85AN19T
C
C    Function  GENPOI : retourne un nombre aleatoire tire selon la
C           loi de POISSON de parametre MT
C---------------------------------------------------------------------
      Integer Function GENPOI(MT)
C--------------------------------------------------------------------
C-----  Generation de nombre pseudo-aleatoires selon une loi de
C-----  Poisson de parametre MT
C--------------- methode de generation loi de Poisson ----------------
C-----  METHODE  = 1/2/3 si (limite centree)/(box-muller)/(rejet)
C-----       METHODEG = 1/2/3 si (FR-1)/(produit)/(gauss)
C-----       GPSTAB        = 0/1/2 stabilisation de la variance
C-----                   (non)/(0.33)/(0.375) cf. EFRON)
C--------------- Attention !!! --------------------------------------
C-----  METHODE=1 et METHODEG=1 non implantees dans ce module car ne
C-----  donnent pas des resultats satisfaisants
C--------------------------------------------------------------------
      Implicit       None
C
      Integer       GPP,GPG
      External       GPP,GPG
C
      Real*4       MT
      Integer*4       Methode,MethodeG,GPStab
C-----
      IF (MT.lt.0.0) THEN
        genpoi=mt
        return
      endif
      IF (MT.lt.30.0) THEN
        METHODE=2
      ELSE IF ((MT.ge.30.0) .AND. (MT.lt.87.0)) THEN
        METHODE=3
        METHODEG=2
        GPSTAB=1
      ELSE IF ((MT.ge.87.0) .AND. (MT.lt.500.0)) THEN
        METHODE=3
        METHODEG=3
        GPSTAB=1
      ELSE IF (MT.ge.500.0) THEN
        METHODE=3
        METHODEG=2
        GPSTAB=0
      END IF
C
      IF (METHODE.eq.2) THEN
        GenPoi = GPP(MT)
      ELSE IF (METHODE.eq.3) THEN
        GenPoi = GPG(MT,MethodeG,GPStab)
      ELSE
        Write(*,*) 'Erreur GenPoi !!'
      END IF
C-----
      Return
      END       ! GenPoi


      Integer FUNCTION GPP(MT)
C----- Loi de Poisson par la methode du produit
C----- Appelle Random
C
C
      Real*4       MT,P,R
      Integer       K
C-----
      P=Random()
      K=0
      R=EXP(-MT)
      DO WHILE (P.gt.R)
        P=P*Random()
        K=K+1
      END DO
      GPP=K
C-----
      Return
      END       ! GPP

      Integer FUNCTION GPG(MT,MethodeG,GPStab)
C----- Loi de Poisson par approx.gauss. (N(0,1) -> P(MT))
C----- Appelle Random,GPP,GPG
C
      Real*4       GG
      External       GG
C
      Real*4       MT
      Integer*4       MethodeG,GPStab,X
C-----
      IF (GPSTAB.le.0) THEN
        X=NINT(MT+SQRT(MT)*GG(MethodeG))
      ELSE IF (GPSTAB.eq.1) THEN
        X=NINT((0.5*GG(MethodeG) + SQRT(MT))**2 - 0.33)
      END IF
      IF (X.lt.0) X=0
      GPG=X
C-----
      Return
      END       ! GPG


      Real*4 FUNCTION GG(MethodeG)
C----- Loi N(0,1) par (lim.centree)/(produit)/(rejet)
C
      Intrinsic       Cos,Log,Sqrt
C
c     Real*4       DPI,SQDSE,SQS12
      Real*4       DPI,SQDSE
      Parameter       (DPI  =6.283185308)       ! 2*PI
      Parameter       (SQDSE=0.857763885)       ! sqrt(2/e)
c     Parameter       (SQS12=0.083333333)       ! sqrt((1/12)/12)
c     Real*4       U,V,S,SIGMA
      Real*4       U,V
      Integer*4       MethodeG
      Logical       SiFin
C-----
      IF (METHODEG.eq.2)THEN
        GG=Sqrt(-2*Log(Random()))*Cos(DPI*Random())
      ELSE IF (METHODEG.eq.3) THEN
        SiFin=.false.
        DO While (.Not.SiFin)
          U=Random()
          V=(2*Random()-1)*SQDSE
          SiFin = ((0.25*V*V/(U*U)).le.-Log(U))
        END DO
        GG=V/U
      END IF
C-----
      Return
      END       ! GG
c----------------------------------------------------------------
      Subroutine FindFmt(Lun,aLine,FMTfields,FMTstring,nC_L,iErr)
C-----------------------------------------------------------------------
C---- The routine "FindFmt" emulates the free format data input
C----         Read(String1,'(a,i,2f,..)') aString,i1,R1,R2,...
C---- but with additional errror checking. Thus, given a description
C---- of the expected fields "FindFmt" returns the format of the line
C---- to be decoded. Valid field descriptors are:
C----          I:integer; R:real; A:free A format; 1 to 5:A1 to A5
C---- Example of use:
C----     Character aLine*(*),FMTfields*(*),FMTstring*(*),String*5
C----     Parameter (iLun=30)       ! input logical unit number
C----
C----  !-- Usual fixed format input (e.g.)
C----     Read(iLun,'(4x,A5,i3,1x,2f8.2,i5)',End=998,Err=999)
C----    +     String,i1,R1,R2,i2
C----
C----  !-- Free format input (Read performed by FindFMT)
C----     FMTfields = '5iffi'
C----     Call FindFmt(iLun,aLine,FMTfields,FMTstring,nC_L,iErr)
C----     if (iErr.eq.-1) GoTo 998 ! End of Line
C----     if (iErr.ne. 0) GoTo 999 ! input error
C----     Read(aLine,FMTstring) String,i1,R1,R2,i2
C----
C----  !-- Free format input (Read performed by calling routine)
C----     Read(iLun,'(a)') aLine
C----     FMTfields = '5iffi'
C----     Call FindFmt(0,aLine,FMTfields,FMTstring,nC_L,iErr)
C----     if (iErr.eq.-1) GoTo 998 ! End of Line
C----     if (iErr.ne. 0) GoTo 999 ! input error
C----     Read(aLine,FMTstring) String,i1,R1,R2,i2
C----     ......
C---- 998 Continue ! End of file encountered
C----     ......
C----  !-- Output error message if any
C---- 999 Call FindFMT_Err(iErr,aLine,nC_L)
C-----------------------------------------------------------------------
C---- A.Filhol ILL Grenoble    24-Mars-92
C-----------------------------------------------------------------------
C---- --> Lun       : Logical Unit Number
C----                  0= aLine must be given
C----                 >0= aLine is read by FindFMT from unit "Lun"
C---- <-> aLine     : character string to be decoded
C---- --> FMTfields : description of the format fields (e.g. IIFIF)
C---- <-- FMTstring : format of the line (e.g. (I5,I1,F8.0,I4,F7.0,) )
C---- <-- nC_L      : number of last character treated in line "aLine"
C---- <-- iErr      : FindFMT error code
C----    -2 : FORTRAN read error
C----    -1 : End of file
C----     0 : No Error
C----     1 : empty format descriptor (0 field)
C----     2 : data string read error
C----     3 : integer field found real !
C----     4 : begged dot, sign or "e" character !
C----     5 : invalid character in an integer field !
C----     6 : invalid field in format descriptor !
C----     7 : invalid character in a numeric field !
C----     8 : 0 character in current field !
C----     9 : format string length exceeded !
C----    10 : separator missing !
C----    11 : incomplete E or D format !
C----    12 : incomplete number !
C------------------------------------------------------------------------
      Implicit    None
c
      Integer*4   LenStr,GetFTMfield
      external    LenStr,GetFTMfield
c
      Character   aLine*(*),FMTfields*(*),FMTstring*(*)
      Integer*4   Lun      ! Logical unit number
      Integer*4   nC_L     ! counts characters in Line
      Integer*4   iErr     ! error code
c
      Integer*4   ioS      ! Fortran status code
      Integer*4   Line_Nb  ! current line number
      Integer*4   L_Fields ! true length of format descriptor
      Integer*4   L_Line   ! true length of data line
      Integer*4   nCar     ! counts characters in current format field
      Integer*4   nFld     ! counts format fields in FMTfields
      Integer*4   nStr     ! counts characters in FMTstring
      Integer*4   iFld     ! field type -1:integer;-2:real;>0:A1 to A9
      Logical     ifSearchEnd
      common/Line_count/Line_Nb
      Include     'fdecode.inc'
c
c---- Initialize
      nC_L = 0
      nFld = 0
      FMTstring = '()'     ! will receive FORTRAN format
      nStr = 1             ! at least a right parentheses in FMTstring
      iErr = iErrNone
      L_Fields  = LenStr(FMTfields)
c
c---- Format descriptor in upper case
      if (FMTfields.eq.' ') then
        iErr = iErrFields           ! empty FMT format descriptor
        Return
      end if
      Call UpperCase(FMTfields)
c
c---- (Get and) verify data line
      if (Lun.gt.0) then
  1      Read(Lun,'(a)',ioStat=ioS,end=2) aLine
      L_Line    = LenStr(aLine)     ! true length without trailing spaces
        if(aLine(1:1).eq.'!'.or.L_line.eq.0) then
         Line_Nb=Line_Nb+1
         goto 1
        endif
  2     if (ioS.eq.-1) then
          iErr = iErrEof            ! End Of File
          Return
        else if (ioS.gt.0) then
          iErr = -ioS-100           ! FORTRAN read error
          Return
        end if
      end if
      if (aLine.eq.' ') then
        iFld = -999                 ! any value but 0 here
        goto 10
      end if
c
c---- Start decoding line character by character
      ifSearchEnd = .false.
c
      Do WHILE (.not.ifSearchEnd)
c
c     -- Get a new format field type
        nCar = 0                    ! new format field
        iFld = GetFTMfield(FMTfields, nFld, L_fields, iErr)
        if (iErr.ne.iErrNone) return! Error in field definition
        if (iFld.eq.iEndFMT) then   ! format exhausted
          if (nFld.eq.0) then
            iErr = iErrInvalField   ! invalid field in FMTfields
            return
          else
            goto 10                 ! scan end
          end if
        end if
c
c     -- Decode current field (character or numeric ?)
        if (iFld.gt.iEndFMT) then
          Call TreatMCharField(iFld,aLine,L_Line,nC_L,nCar,iErr)
        else if (iFld.eq.iEndFMT) then    ! format exhausted
          goto 10
        else if (iFld.lt.iEndFMT) then
          Call TreatNumerField(iFld,aLine,L_Line,nC_L,nCar,iErr)
        end if
        if (iErr.ne.iErrNone) Return
        if ((iFld.lt.iEndFMT.and.nCar.eq.0) .or. iFld.eq.0) then
          iErr = iErrEmptyField           ! no characters in field
          Return
        end if
c
c     -- Build current FMT element
        Call BuildFMT(iFld,nCar,nStr,FMTstring,iErr)
        if (iErr.ne.iErrNone) return      ! format string length exceeded
c
c     -- End of data Line ?
        if (nC_L.ge.L_Line) ifSearchEnd = .true.
c
      END DO
c
c---- Terminates and close the format field
   10 Continue
c     -- If FMT not exhausted we append the remaining fields to
c     -- the format string
      if (iErr.eq.iErrNone .and. nFld.lt.L_Fields) then
        do while (iFld.ne.iEndFMT)
          iFld = GetFTMfield(FMTfields, nFld, L_fields, iErr)
          if (iErr.ne.iErrNone) return    ! Error in field definition
          if (iFld.ne.iEndFMT) then
            Call BuildFMT(iFld,1,nStr,FMTstring,iErr)
            if (iErr.ne.iErrNone) return  ! format string length exceeded
          end if
        end do
      end if
c     -- Close format string
      FMTstring(nStr:nStr) = ')'
c
c---- End
      Return
      End   ! FindFmt




      Integer*4 Function GetFTMfield(FMTfields,nFld,nFldMax,iErr)
C-----------------------------------------------------------------------
C---- Get current field type
C----  --> FMTfields: format descriptor
C----  <-> nFld   : current field in format descriptor
C----  --> nFldMax: max. number of fields in format descriptor
C----  <-- iErr   : error return (unchanged or "iErrInvalField")
C----  <-- Function value :
C----    iInte  : Integer field "I"
C----    iReal  : real field "F"
C----    0 to 9 : fixed format A1 to A9 field
C----    iEndFMT: format exhausted
C-----------------------------------------------------------------------
      Implicit    None
      Integer*4   nFld,nFldMax,iErr
      Character   FMTfields*(*),Car*1
      Include     'fdecode.inc'
c
        nFld = nFld + 1
        if (nFld.gt.nFldMax) then
          GetFTMfield = iEndFMT
        else
          Car = FMTfields(nFld:nFld)
          if (Car.eq.'I') then
            GetFTMfield = iInte
          else if (Car.eq.'F') then
            GetFTMfield = iReal
          else if (iChar(Car).ge.i_One .and. iChar(Car).le.i_Nine) then
            GetFTMfield = iChar(Car)
          else
            GetFTMfield = iEndFMT
            iErr = iErrInvalField         ! Error in field definition
          end if
        end if
c
      Return
      End   ! GetFTMfield



      Subroutine TreatMCharField(iFld,aLine,L_Line,nC_L,nC_X,iErr)
C-----------------------------------------------------------------------
C---- Fixed length "A1 to A9" field : A<iFld-48>
C---- Leading spaces are ignored; separators : space and Tab
C----  <-> iFld   : "A" format size (1 to 9)
C----  --> aLine  : data line to be analysed
C----  --> L_Line : true length of data line
C----  <-> nC_L   : Current character in data line
C----  <-- nC_X   : number of characters in X format field
C----               n = Char(iFld); m=nC_X
C----               1st format field: An;  others: mX,An;
C----  <-- iErr   : error code
C-----------------------------------------------------------------------
      Implicit    None
c
      Character   aLine*(*)
      Integer*4   iFld   ! 1 to 9 for "A1 to A9" FORTRAN format
      Integer*4   L_Line ! length of data Line
      Integer*4   nC_L   ! current character in data line
      Integer*4   nC_X   ! size of the X FORTRAN format field
      Integer*4   iErr   ! error code
c
      Character   Car*1
      Integer     nCar
      Logical     ifEnd
      Character       cTab*1
      Include     'fdecode.inc'
      cTab=Char(9)
c
      nC_X = 0
      iErr = 0
c
c---- End of ligne
      if (nC_L.ge.L_Line) return
c
c---- if not 1rst field, 1rst character must be a separator
      if (nC_L.gt.1) Then
        nC_L = nC_L+1
        Car  = aLine(nC_L:nC_L)
        if (Car.ne.' ' .and. Car.ne.cTab) then
          iErr = iErrSepMiss              ! separator missing
          Return
        end if
        nC_X = nC_X+1
      End if
c
c---- Remove leading spaces
      ifEnd = .false.
      Do While (.not.ifEnd)
        if (nC_L.ge.L_Line) return        ! end of line
        nC_L = nC_L+1
        Car  = aLine(nC_L:nC_L)
        if (Car.eq.' ') Then
          nC_X = nC_X+1                   ! count leading spaces
        else
          ifEnd = .true.                  ! 1rst valid character
          nC_L = nC_L-1
        end if
      End do
c
c---- Count characters until next Tab or end of line
      nCar = 0
      ifEnd = .false.
      Do While (.not.ifEnd)
        if (nC_L.lt.L_Line .and. nCar.lt.(iFld-48)) then
          nC_L = nC_L+1
          nCar = nCar+1
          Car = aLine(nC_L:nC_L)
          if (Car.eq.' ' .or. Car.eq.cTab) then
            ifEnd = .true.                ! separator found
            nCar  = nCar - 1
            nC_L  = nC_L - 1
          end if
        else
          ifEnd = .true.                  ! end of line
        end if
      End Do
c
c---- Load size of the A format field
      if (nCar.eq.0) Then
        iErr = iErrEmptyField             ! no charac. in field
      else
        iFld = nCar+48                    ! true size of the A field
      end if
c
      Return
      end   ! TreatMCharField



      Subroutine TreatNumerField(iFld,aLine,L_Line,nC_L,nCar,iErr)
C-----------------------------------------------------------------------
C----  Free "I" and "F" formats
C----  Look for a separator (space or Tab) after any valid character
C----
C----  --> iFld   : Format file type (1 to 9)
C----  --> aLine  : data line
C----  --> L_Line : true length of data line
C----  <-> nC_L   : Current character in data line
C----  <-- nCar   : number of characters in format field
C----  <-- iErr   : error code
C-----------------------------------------------------------------------
      Implicit    None
c
      Integer*4   iFld  ! field type
      Character   aLine*(*)
      Integer*4   L_Line! true length of the data line
      Integer*4   nCar  ! counts characters in format field
      Integer*4   nCar1 ! 1st usefull character in field
      Integer*4   nC_L  ! counts characters in data line
      Integer*4   iErr  ! error code
c
      Integer*4   nPosi ! number of 1st character in field
      Character   Car*1,Car_*1
      Logical     ifEnd,ifDot,ifSign
      Character       cTab*1
      Include     'fdecode.inc'
      cTab=Char(9)
c
      iErr   = 0
      nCar   = 0
      ifDot  = .false.
      ifSign = .false.
      nPosi  = nC_L
c
c---- Skip previous separator (space, Tab or sign) and leading spaces
      ifEnd = .false.
      Do While (.not.ifEnd)
        nC_L = nC_L+1
        if (nC_L.le.L_Line) then
          nCar = nCar+1
          Car = aLine(nC_L:nC_L)
c         -- Tab character
          if (Car.eq.cTab) Then
            if (nCar.eq.1 .and. nC_L.gt.1) then
              aLine(nC_L:nC_L) = ' '      ! previous separator
            else
              if (ifSign) then
                iErr = iErrNumber         ! incomplete number
                return
              end if
              nC_L = nC_L-1               ! new separator
              nCar = nCar-1
              return
            end if
c         -- a sign
          else if (Car.eq.'+' .or. Car.eq.'-') then
            ifSign = .true.
c         -- a space
          else if (Car.eq.' ') then
            if (ifSign) then
              iErr = iErrNumber           ! incomplete number
              return
            end if
c         -- any other character
          else
            ifEnd = .true.
          end if
        else
          Return                          ! end of line
        end if
      End Do
c
c     -- No valid previous separator found (Except for 1st field)
      if (nPosi.gt.1 .and. nCar.eq.1) then
        iErr = iErrSepMiss                ! separator missing
        Return
      end if
c
c---- Check first character and initialize search
c
c     -- Decimal point -> valid in real fields only
      if (Car.eq.'.') then
        ifDot = .true.
        if (iFld.ne.iReal)  then
          iErr = iErrFieldType            ! not an integer field
          Return
        end if
c     -- e,E,d,D -> always invalid at this position
      else if(Car.eq.'E'.or.Car.eq.'e'.or.Car.eq.'d'.or.Car.eq.'D') then
        if (iFld.eq.iReal) then
          iErr = iErrEfrmt                ! incomplete E or D format
        else
          iErr = iErrInvalC               ! invalid char in int. field
        end if
        return
c     -- invalid if not a sign or a digit
      else if (iChar(Car).lt.i_Zero .or. iChar(Car).gt.i_Nine) then
        iErr = iErrInvalChar        ! invalid character
        return
      end if
c
c---- save position of first character
      nCar1 = nCar

c
c---- Count characters in number
      ifEnd = .false.
      Do While (.not.ifEnd)
c
        Car_ = Car      ! save previous character
c
        nC_L = nC_L+1
        if (nC_L.le.L_Line) then
          nCar = nCar+1
          Car = aLine(nC_L:nC_L)
c       -- Current character is a decimal point
          if (Car.eq.'.') then
            if (ifDot) then
              iErr = iErrCharBegg         ! begged character (dot)
              Return
            else if (iFld.ne.iReal) then
              iErr = iErrFieldType        ! not an integer field
              Return
            else
              ifDot = .true.
            end if
c       -- Current character is a space or Tab (separator)
          else if (Car.eq.' ' .or. Car.eq.cTab) then
            if (Car_.eq.'+' .or. Car_.eq.'-') then
              iErr = iErrNumber           ! incomplete number
              return
            end if
            ifEnd = .true.
            nCar  = nCar - 1
            nC_L  = nC_L - 1
c       -- Current character is a sign
          else if (Car.eq.'+' .or. Car.eq.'-') then
            if (Car_.eq.'+' .or. Car_.eq.'-') then
              iErr = iErrCharBegg         ! begged character
              return
            else if (nCar.gt.nCar1) then
              if (iFld.eq.iReal) then
                if (Car_.ne.'E' .and. Car_.ne.'e' .and.
     +              Car_.ne.'D' .and. Car_.ne.'d') then
                  ifEnd = .true.          ! Sign is a valid separator
                  nCar  = nCar - 1
                  nC_L  = nC_L - 1
                  Return
                end if
              else                        ! Sign is a valid separator
                ifEnd = .true.
                nCar  = nCar - 1
                nC_L  = nC_L - 1
                Return
              end if
            end if
c       -- Current character is a "e E d D"
          else if (Car.eq.'E'.or.Car.eq.'e' .or.
     +             Car.eq.'d'.or.Car.eq.'D') then
            if (nCar.eq.nCar1 .or. Car_.eq.'-'.or.Car_.eq.'+') then
              iErr = iErrEfrmt            ! incomplete E or D format
              return
            else if (Car_.eq.Car) then
              iErr = iErrCharBegg         ! begged character
              return
            end if
c       -- Ccurrent character is not a valid one ?
          else if (iChar(Car).lt.i_Zero .or. iChar(Car).gt.i_Nine) then
            iErr = iErrInvalChar          ! invalid character
            Return
          end if
        else
          ifEnd = .true.                  ! end of line
        end if
c
      End do
c
      Return
      End   ! TreatNumerField


      Subroutine BuildFMT(iFld,nCar,nStr,FMTstring,iErr)
C-----------------------------------------------------------------------
C---- Add a new field to the FMT string
C----  -> iFld     : format type
C----  -> nCar     : integer or real field :
C----                  nber of characters in field
C----              : character field :
C----                  nber of characters to skip before A field
C---- <-> nStr     : current character number in FMTstring
C---- <-> FMTstring: FORTRAN format string
C---- <-  iErr     : Error code
C-----------------------------------------------------------------------
      Implicit    None
c
      Integer*4   iFld,nCar,nStr,iErr
      Character   FMTstring*(*)
c
      Integer*4   N
      Include     'fdecode.inc'
c
c---- heading symbol "F"
      nStr = nStr + 1
      if (nStr.gt.Len(FMTstring)) goto 10
      if (iFld.eq.iInte) then
        FMTstring(nStr:nStr)  = 'I'
      else if (iFld.eq.iReal) then
        FMTstring(nStr:nStr)  = 'F'
      else if (iFld.gt.0) then
        if (nCar.eq.0) then
          FMTstring(nStr:nStr)  = 'A'
        else
          FMTstring(nStr:nStr+3)  = Char(nCar+48)//'x,A'
          nStr = nStr + 3
        end if
      end if
c
c---- numeric part of Integer and Real fields
      if (iFld.lt.0) then
c     -- hundredth
        if (nCar.ge.100) then
          N = Int(nCar/100)
          nStr = nStr + 1
          if (nStr.gt.Len(FMTstring)) goto 10
          FMTstring(nStr:nStr) = Char(N+48)
          nCar = nCar - N*100
        end if
c     -- tenth
        if (nCar.ge.10) then
          N = Int(nCar/10)
          nStr = nStr + 1
          if (nStr.gt.Len(FMTstring)) goto 10
          FMTstring(nStr:nStr) = Char(N+48)
          nCar = nCar - N*10
        end if
c     -- units
        nStr = nStr + 1
        if (nStr.gt.Len(FMTstring)) goto 10
        FMTstring(nStr:nStr) = Char(nCar+48)
c     -- Add ".0" to the end of Real fields
        if (iFld.eq.iReal) then
          nStr = nStr + 2
          if (nStr.gt.Len(FMTstring)) goto 10
          FMTstring(nStr-1:nStr) = '.0'
        end if
c
c---- numeric part of "A" fields
      else if (iFld.gt.0) then
        nStr = nStr + 1
        if (nStr.gt.Len(FMTstring)) goto 10
        FMTstring(nStr:nStr)   = Char(iFld)
      end if
c
c---- Add a separator "," after each new FORTRAN field
      nStr = nStr + 1
      if (nStr.gt.Len(FMTstring)) goto 10
      FMTstring(nStr:nStr) = ','
c
      Return
c
c---- Error
   10 iErr = iErrStrLength          ! format string length exceeded
      Return
      End   ! BuildFMT



      Subroutine FindFMT_Err(iErr,aLine,nC_L)
C-----------------------------------------------------------------------
C----- Output the error messages from FindFMT
C-----  iErr  : error Code
C-----  aLine : Current data line
C-----  nC_L  : location of last character treated
C-----------------------------------------------------------------------
      Implicit    None
c---- Error messages --------
       Integer*4       MssgBeg,MssgEnd
       Parameter       (MssgBeg=-2)              ! lower message number
       Parameter       (MssgEnd=12)              ! upper message number
       Character*80 Message(MssgBeg:MssgEnd)
       data Message/
     + 'FindFMT: data line FORTRAN read error nber: ',
     + 'FindFMT: End of file !',
     + 'FindFMT: no error',
     + 'FindFMT: empty format descriptor (0 field) !',
     + 'FindFMT: data string, read error !',
     + 'FindFMT: integer field found real !',
     + 'FindFMT: begged dot, sign or "e" character !',
     + 'FindFMT: invalid character in an integer field !',
     + 'FindFMT: invalid field in format descriptor !',
     + 'FindFMT: invalid character in a numeric field !',
     + 'FindFMT: 0 character in current field !',
     + 'FindFMT: format string length exceeded !',
     + 'FindFMT: separator missing !',
     + 'FindFMT: incomplete E or D format !',
     + 'FindFMT: incomplete number !'/
c
      integer*4   LenStr
      External    LenStr
c
      Character   aLine*(*),auxi*256
      Integer*4   iErr,nC_L
c
      Integer*4   Ln
      Character   LaMarque*40
      Include     'fdecode.inc'
c
c---- Error message
      if (iErr.eq.iErrNone .or. iErr.eq.iErrEof) then
        Return
      else if (iErr.lt.iErrEof) then
        Write(*,'(1x,a,i4)') Message(-2)(1:LenStr(Message(-2))),
     +                      -(iErr+100)
      else if (iErr.lt.MssgBeg .or. iErr.gt.MssgEnd) then
        Write(*,'(1x,a,i2)') 'FMT decode error nber:',iErr
      else
        Write(*,'(1x,a)') Message(iErr)(1:LenStr(Message(iErr)))
      end if
c
c---- Output data line and print a mark at error location
      Ln = max0(LenStr(aLine),1)
      if (Ln.le.129) then
        auxi='"'//aLine(1:Ln)//'"'
        Write(*,'(1x,a)') auxi
        if (nC_L.eq.1) then
          Write(*,'(1x,6h ^----/)')
        else if (nC_L.gt.1) then
          Write(LaMarque,'(a,i3,a)')  '(1x,', nC_L, '(1h-),''^''/)'
          Write(*,LaMarque)
        end if
      else
        Write(*,'(1x,a)') aLine(1:Ln)
        Write(LaMarque,'(a,i3,a)')  '(1x,', nC_L-1, '(1h-),''^''/)'
        Write(*,LaMarque)
      end if
c
      Return
      End   ! FindFMT_Err


      Subroutine UpperCase(String)
C-----------------------------------------------------------------------
C---- Convert string to uppercase
C-----------------------------------------------------------------------
      Implicit    None
      Character   String*(*)
      Integer     i,Ln,iCar
c
      Ln = Len(String)
      do i=1,Ln
        iCar = iChar(String(i:i))
        if (iCar.ge.97 .and. iCar.le.122) String(i:i) = Char(iCar-32)
      end do
c
      return
      end   ! UpperCase
c----------------------------------------------------------------------
      Subroutine LowerCase(String)
C-----------------------------------------------------------------------
C---- Convert string to Lowercase
C-----------------------------------------------------------------------
      Implicit    None
      Character   String*(*)
      Integer     i,Ln,iCar
c
      Ln = Len(String)
      do i=1,Ln
        iCar = iChar(String(i:i))
        if (iCar.ge.65 .and. iCar.le.90 ) String(i:i) = Char(iCar+32)
      end do
c
      return
      end   ! LowerCase


      Integer*4 Function LenStr(String)
C-----------------------------------------------------------------------
C---- Returns true length of "String" (assumes "TRIM" not available)
C-----------------------------------------------------------------------
      Implicit    None
      Character   String*(*)
      integer*4   i

        LenStr = 0
        do i=LEN(String),1,-1
          if (String(i:i).ne.' ') then
            LenStr = i
            Return
          end if
        end do
      Return
      End   ! LenStr
c------------------------------------------------------------------------
C
C***            Prepare output file **.igr [text input for Igor]
C
      subroutine Wigor(icz)
      include 'ful0.inc'
      include 'ful1.inc'
      include 'ful2.inc'
      COMMON/G2/S1,S2,S3,D1,D2,R1,R2,R3,DEVINC,YINV,DW,QQ1,QQ2,QQ3,SCOR
c
      Common/Rfactor/freed,freed1,chi2,cchi2
c
      Character*240 tphas
      Character*6 Titlph(8)
c       Character*6 Titlph(8),AlrtStr*40
c       Character jour*9,heure*9
      character idat*8,itim*11
c       common/diahora/jour,heure
cSun
c      character idat*24
c      common/diahora/idat
cSun
      common/diahora/idat,itim
      Dimension dif(IDSZ),indx(irs),plin(8,irs)
      igor=i_prf
      yMaxi=0.0
      dimin=99999.
      dimax=-99999.
      do j=1,npts
        if(txv(j).lt.thmin.or.txv(j).gt.thmax)goto 8002
        if(iabs(jobtyp).le.1)then
                dif(j)=y(j)-yc(j)
                if(dif(j).lt.dimin)dimin=dif(j)
                if(dif(j).gt.dimax)dimax=dif(j)
                if(y(j).gt.ymaxi)ymaxi=y(j)
        endif
        if(yc(j).gt.ymaxi)ymaxi=yc(j)
8002  end do
      ltik=int(ymaxi/30.)       ! Size of line position tick marks
      ispac=ltik/4              ! Spacing tick to diff & diff to axis
      ishft=-int(nphase*(ispac+ltik)+dimax)     ! Offset of difference curve
C     if(Nexcrg.gt.0)ishft=ishft-ltik   !Add space for excluded region markers
      ymaxi=100.*int(ymaxi/100.)                ! Roundoff intensity scale
c     if(iabs(jobtyp).le.1)then
c     ymin=-2*ispac-int(nphase*ltik+dimax+abs(dimin))   ! Lowest bound of y axis
c      else
c     ymin=-ispac-int(nphase*ltik)
c     endif
      lex=-int(nphase*ltik+ispac*(nphase-1))            !       Excluded regions
      ltik=-ltik                                        ! Plot ticks below zero
      ilt=LenStr(phsnm(1))
      if(ilt.gt.30)ilt=30
      tphas=phsnm(1)(1:ilt)
       do k=2,nphase
        il=LenStr(phsnm(k))
        if(il.gt.30)il=30
        tphas=tphas(1:ilt)//'\r'//phsnm(k)(1:il)   ! concatenate phase names
        ilt=ilt+il+2
       end do
C
C--- Find phase number for each reflection (identify line markers)
C
      Do i=1,nphase
      indx(i)=0                         ! Index of reflections for phase i
      write(Titlph(i),'(''Phase'',i1)')i ! Name of reflection list for Igor
      end do
      Do j=1,icz
      jtik=int(irefs(j)/256/256/256/8)  ! Decode phase number
      indx(jtik)=indx(jtik)+1           ! count reflections of phase jtik
      plin(jtik,indx(jtik))=refs(j,2)+glb(1) ! renumber lines & add zeroshift
      end do
C
C--- Write waves (Iobs,Icalc, Diff & line positions for every phase)
C
      il=LenStr(tphas)
      write(igor,3000)
3000  format('IGOR')
      if(iabs(jobtyp).le.1)then
      write(igor,3001)
3001  format('WAVES     TwoTheta,       Iobs,   Icalc,  Diff'/'BEGIN')
      write(igor,3010)(txv(j),y(j),yc(j),dif(j),j=1,npts)
3010  format(f7.3,3f10.0)
        else
      write(igor,3002)
3002  format('WAVES     TwoTheta,       Icalc'/'BEGIN')
      write(igor,3012)(txv(j),yc(j),j=1,npts)
3012  format(f7.3,f10.0)
      endif
      write(igor,3020)
3020  format('END')
      do i=1,nphase
      write(igor,3100)Titlph(i),i
3100  format('WAVES     ',a6,', tik',i1/'BEGIN')
      write(igor,3101) (plin(i,j),ltik,j=1,indx(i))
3101  format(f7.3,1x,i6)
      write(igor,3020)
      end do
C
C--- Add excluded regions as additional "waves"
C
      if(Nexcrg.gt.0)then
      do i=1,Nexcrg
       if(alow(i).lt.thmin)then
       xlow=thmin
        else
       xlow=alow(i)
       endif
       if(ahigh(i).gt.thmax)then
       xhigh=thmax
       else
       xhigh=ahigh(i)
       endif
       if(i.le.9)then
       write(igor,3150)i,i
3150   format('WAVES Excrg',i1,', excl',i1/'BEGIN')
       write(igor,3151)xlow,lex,xhigh,lex
3151   format(f8.2,1x,i6/f8.2,1x,i6/'END')
        else
       write(igor,3152)i,i
3152   format('WAVES Excrg',i2,', excl',i2/'BEGIN')
       write(igor,3153)xlow,lex,xhigh,lex
3153   format(f8.2,1x,i6/f8.2,1x,i6/'END')
       endif
       end do
       endif
C
C--- Set up plot parameters
C
       write(igor,3200)
3200   format( 'X Silent 1'/'X DefaultFont Times')
C
C--- Now plot profiles and set x scale
C
       if(iabs(jobtyp).le.1)then
       write(igor,3300)thmin,thmax,ishft
3300   format('X Display Iobs, Icalc, Diff vs TwoTheta'/
     1'X Modify mode(Iobs)=2, rgb(Iobs)=(0,0,0),',
     2' rgb(Diff)=(3800,44400,4000)'/
     3'X SetAxis bottom ',f8.2,', ',f8.2/
     4'X Modify lsize(Iobs)=2'/'X Modify offset(Diff)={0,',i7,'}')
       else
       write(igor,3302)
3302  format('X Display Icalc vs TwoTheta')
      endif
      write(igor,3304)
3304  format('X Modify wbRGB=(48483,64909,65535),',
     3' gbRGB=(64612,65535,49415)'/
     4'X Modify mirror(left)=2, mirror(bottom)=2'/
     5'X Modify gfSize=12')
C
C--- then add excluded regions in "fill to zero" mode
C
      if(Nexcrg.gt.0)then
       do i=1,Nexcrg
         if(i.le.9)then
       write(igor,3370)i,i,i,i,i
3370   format('X Append excl',i1,' vs Excrg',i1/
     1'X Modify mode(excl',i1,')=7, hbFill(excl',i1,')=4'/
     2'X Modify rgb(excl',i1,')=(6421,62320,64794)')
      else
      write(igor,3371)i,i,i,i,i
3371  format('X Append excl',i2,' vs Excrg',i2/
     1'X Modify mode(excl',i2,')=7, hbFill(excl',i2,')=4'/
     2'X Modify rgb(excl',i2,')=(6421,62320,64794)')
      endif
      end do
      endif
C
C--- then plot line markers
C
      do i=1,nphase
      write(igor,3350)i,Titlph(i),i
3350  format('X Append tik',i1,' vs ',a6/'X Modify mode(tik',i1,')=1')
      end do
      do i=2,nphase
      ishtik=(ltik-ispac)*(i-1)
      write(igor,3360)i,ishtik
3360  format('X Modify offset(tik',i1,')={0,',i6,'}')
      end do
C
C--- Add title box
C
          il = LenStr(Filenam)
      if(iabs(jobtyp).le.1)then
      write(igor,3400)filenam(1:il),idat,itim,tphas(1:ilt),chi2
3400    format('X TextBox /F=2/S=3/A=RT "Files: ',a,
     1  '\rDate of fit: \Z09',a9,'/',a8,'\Z12\r',a,'\rChi2 = ',f7.2,'"')
c       write(igor,3400)tphas(1:ilt),idat,itim,chi2
c3400   format('X TextBox /F=2/S=3/A=RT "',a,'\rDate of fit: \Z09',a9,'/',
c     1 a8,'\Z12\rChi2 = ',f7.2,'"')
      else if(idum.eq.0) then
                write(igor,3402)filenam(1:il),idat,itim,tphas(1:ilt)
3402    format('X TextBox /F=2/S=3/A=RT "File:',a,'
     1  \rPattern calculation: \Z09',a9,'/',a8,'\Z12\r',a,'"')
c       write(igor,3402)idat,itim,tphas(1:ilt)
c3402   format('X TextBox /F=2/S=3/A=RT "Pattern calculation: \Z09',a9,'/'
c     1 ,a8,'\Z12\r',a,'"')
      else
       write(igor,3403)filenam(1:il),idat,itim,tphas(1:ilt)
3403    format('X TextBox /F=2/S=3/A=RT "File:',a,'
     1  \rPattern calculation: \Z09',a9,'/',
     2a8,'\Z12\rRandomized with Poisson noise\r',a,'"')
c      write(igor,3403)idat,itim,tphas(1:ilt)
c3403  format('X TextBox /F=2/S=3/A=RT "Pattern calculation: \Z09',a9,'/'
c     1,a8,'\Z12\rRandomized with Poisson noise\r',a,'"')
      endif
      write(igor,3450)
3450  format('X Label left "\Z14Intensity"'/
     1'X Label bottom "\Z142-Theta (in degrees)"'/
     2'X TileWindows/O=1/C'/'X ShowInfo')
C
C--- Finally add comments at the end of Igor file
C
      il=LenStr(Title)
      write(igor,3500)Title(1:il)
3500  format('X | Title of data set: ',a)
      write(igor,3410)idat,itim
c     write(igor,3410)Jour,Heure
3410  format('X | Date of run:',a9,' / ',a8/'X Beep'/'X Beep')
      close(unit=igor)
C
C***    End of Igor part
C
       return
       end
c------------------------------------------------------------------------
c------------------------------------------------------------------------
C SUBROUTINE TO WRITE FILES "LNS.PLT1" AND "LNS.PLT2",
C READY TO BE PLOTED WITH THE LNS SYSTEM
C---------------------------------------------------------------------
      SUBROUTINE PICSURE(ICZ,ymini,ymaxi)
      parameter (i_pl1=1,i_pl2=20)
      INCLUDE 'ful0.inc'
      INCLUDE 'ful1.inc'
      INCLUDE 'ful2.inc'
C
C---------------------------------------------------------------------
C
      if(ymaxi-ymini.le.0.) return
      L=LenStr(filenam)
      OPEN(UNIT=i_pl1,FILE=FILENAM(1:L)//'.PL1',STATUS='unknown')
      OPEN(UNIT=i_pl2,FILE=FILENAM(1:L)//'.PL2',STATUS='unknown')
         DO I=1,NEXCRG
          if(iconstep.eq.1) then
           NPTLOW=(ALOW(I)-THMIN)/STEP+1.5
           NPTHI=(AHIGH(I)-THMIN)/STEP+1.5
          else
           call locate(npts,txv,alow(i),nptlow)
           call locate(npts,txv,ahigh(i),npthi)
          endif
            IF(NPTLOW.LT.1) NPTLOW=1
            IF(NPTLOW.GT.NPTS) NPTLOW=NPTS
            IF(NPTHI.LT.1) NPTHI=1
            IF(NPTHI.GT.NPTS) NPTHI=NPTS
            DO J=NPTLOW,NPTHI
            YC(J)=0.0
            Y(J)=0.0
            END DO
         END DO
C--------------------------------------------------------------------
C
C   WRITTING DATA IN ***.PLT1 AND ***.PLT2
C
C-------------------------------------------------------------------
C
      DO I=1,NPTS
      TWOTH=Txv(i)
      YDIF=Y(I)-YC(I)
      WRITE(i_pl1,*)TWOTH,Y(I),YC(I),YDIF
      ITWOTH=INT(10.*TWOTH)
          DO J=1,NEXCRG
              AL=ALOW(J)-STEP
              AH=AHIGH(J)+STEP
              IL=INT(10.*(ALOW(J)-STEP))
              IAHIGH=INT(10.*(AHIGH(J)))
              IF (IL.EQ.ITWOTH) THEN
                WRITE(i_pl1,*)AL,0.0,0.0,0.0
              ELSE IF (ITWOTH.EQ.IAHIGH) THEN
                WRITE(i_pl1,*)AH,0.0,0.0,0.0
              ENDIF
           END DO
      END DO
C
          AMARK2=0.8
      DO I=1,ICZ
       SHIFT = GLB(8) * COS(REFS(I,2)/2./57.2958)
     &       + GLB(9) * SIN(REFS(I,2)/57.2958)
        ZERO=GLB(1)+SHIFT
      ITICK=IREFS(I)/140000000
      AMARK1=10.0-ITICK
      JREFS=INT(10.*REFS(I,2))
C
          DO J=1,NEXCRG
          ILL=INT(10.*ALOW(J))
          IHH=INT(10.*AHIGH(J))
                IF (JREFS.GT.ILL.AND.JREFS.LT.IHH) GOTO 80
          ENDDO
C
      WRITE(i_pl2,*)REFS(I,2)+ZERO,AMARK1,AMARK2
80    END DO
      RETURN
      END
c------------------------------------------------------------------------
C This subroutine is mainly for Sun machines, for easy use of the
C graphic inteface provided by: Xvgr
C---------------------------------------------------------------------
      SUBROUTINE Plot_XVGR(ICZ,ymini,ymaxi)
      INCLUDE 'ful0.inc'
      INCLUDE 'ful1.inc'
      INCLUDE 'ful2.inc'
      parameter (i_pl1=1,i_pl2=2)
C
C---------------------------------------------------------------------
C
      L=LenStr(filenam)
      OPEN(UNIT=i_pl1,FILE=FILENAM(1:L)//'.plt1',STATUS='unknown')
      OPEN(UNIT=i_pl2,FILE=FILENAM(1:L)//'.plt2',STATUS='unknown')
         iShb=int(ymini/4.0)
         iShd=int(ymaxi/4.0)
         ideltr=int(ymaxi/16)
      do  i=1,npts
          twtet=txv(i)
          idd=int(y(i)-yc(i))
           DO  J=1,NEXCRG
            IF( TWTET.GE.ALOW(J).AND.TWTET.LE.AHIGH(J) ) idd=0
           end do
          write(i_pl1,'(f8.4,4(1x,i8))')
     +      twtet,int(y(i)),int(yc(i)),idd-iShd,int(bk(i))-iShb
      end do
      do j=1,icz
       SHIFT = GLB(8) * COS(REFS(j,2)/2./57.2958)
     }      + GLB(9) * SIN(REFS(j,2)/57.2958)
        ZERO=GLB(1)+SHIFT
        IRL=MOD(IREFS(j),256)-128
        IRK=MOD(IREFS(j)/256,256)-128
        IRH=MOD(IREFS(j)/256/256,256)-128
        twtet=refs(j,2)+zero
        k= IREFS(j)/256/256/256/8
        iposr=-(k-1)*ideltr
        write(i_pl2,'(f8.4,i8,a,3i3,a,i3)')
     +   twtet,iposr,' (',irh,irk,irl,') ',ihkl(j)
       end do
      RETURN
      END
c------------------------------------------------------------------------
c------------------------------------------------------------------------
